use std::path::PathBuf;
use clap::Parser;
use std::fs::File;
use std::io::prelude::*;
use std::num::NonZeroU32;
use std::rc::Rc;
use pixels::{Error, Pixels, SurfaceTexture};
use tao::{
    event::{Event, KeyEvent, WindowEvent, ElementState},
    event_loop::{ControlFlow, EventLoop},
    window::WindowBuilder,
    keyboard::KeyCode,
};
use tao::dpi::LogicalSize;
use tracing::{info, error};
use tracing_subscriber::{layer::SubscriberExt, util::SubscriberInitExt};
use std::sync::{Arc, Mutex};
use rand::prelude::*;
use tao::keyboard::Key;
use std::time::{Duration, Instant};
use cpal::{SampleFormat, SampleRate, Stream, SupportedStreamConfig};
use cpal::traits::{DeviceTrait, HostTrait, StreamTrait};
use rand::seq::index::sample;
use std::sync::mpsc::{sync_channel, Receiver, SyncSender};

const MODERNSHIFT: bool = false;
const OLDJUMP: bool = true;

const MODERNLOADSAVE: bool = false;

#[derive(Parser, Debug)]
struct Args {
    #[arg(short, long)]
    file: PathBuf,
}

struct AudioPlayer {
    stream: Stream,
}

struct Oscilator {
    sample_index: f32,
    volume: f32,
}

impl AudioPlayer {
    fn make(rx: Receiver<f32>) -> Self {
        let host = cpal::default_host();
        let device = host
            .default_output_device()
            .expect("no output device available");

        let config = device.supported_output_configs().expect("error while querying config").find(|config| config.sample_format() == SampleFormat::F32);
        if config.is_none() {panic!("no supported config found");}
        let config = config.unwrap();
        if config.min_sample_rate() > SampleRate(48000) || config.max_sample_rate() < SampleRate(48000) {
            //("{:?},{:?}", config.min_sample_rate(), config.max_sample_rate());
            panic!("sample rate out of range");
        }

        let format = SupportedStreamConfig::new(
            config.channels(),
            SampleRate(48000),
            config.buffer_size().clone(),
            SampleFormat::F32
        );

        //let time_at_start = Instant::now();
        let mut oscilator = Oscilator{sample_index: 0.0, volume: 0.0};
        let stream = device.build_output_stream(
            &format.config(),
            move |data: &mut [f32], _: &cpal::OutputCallbackInfo| {
                //let time_since_start = Instant::now().duration_since(time_at_start).as_secs_f32();
                if let Ok(vol) = rx.try_recv() {
                    oscilator.volume = vol as f32;
                }
                process_frame(data, config.channels().into(), &mut oscilator);
            },
            (|err| error!("Error while creating output stream, {}", err)),
            None,
        );

        Self {
            stream: stream.expect("no stream available"),
        }

    }

    fn run(&mut self) {
        self.stream.play().unwrap();
    }
}

fn process_frame(output: &mut [f32], channels: usize, oscilator: &mut Oscilator) {
    for frame in output.chunks_mut(channels) {
        let value = oscilator.tick() * oscilator.volume;

        for sample in frame.iter_mut() {
            *sample = value;
        }
    }
}

impl Oscilator {
    fn tick(&mut self) -> f32 {
        self.sample_index = (self.sample_index + 1.0) % 48000.0;
        let two_pi = 2.0 * std::f32::consts::PI;
        (self.sample_index * 440.0 * two_pi/48000.0).sin()
    }
}

#[derive(Copy, Clone)]
struct Chip8 {
    display: [u64; 32],
    index_register: u16,
    registers: [u8; 16],
    sound_timer: u8,
    delay_timer: u8,
    ram: [u8; 4096],
    pc: u16,
    stack: [u16; 64],
    stack_pointer: u8,
    keys: [bool; 16],
    get_key_pause: bool,
    keys_released: [bool; 16],
    keys_pressed: [bool; 16],
    display_wait: bool,
}

impl Chip8 {
    fn new() -> Self {
        Chip8 {
            display: [0;32],
            index_register: 0,
            registers: [0;16],
            sound_timer: 0,
            delay_timer: 0,
            ram: [0; 4096],
            pc: 0x200,
            stack: [0; 64],
            stack_pointer: 0,
            keys: [false; 16],
            get_key_pause: false,
            keys_released: [false; 16],
            keys_pressed: [false; 16],
            display_wait: false,
        }
    }

    fn insert_font(mut self) -> Self {
        font_initialiser(&mut self.ram);
        self
    }

    fn load_file(mut self, file: PathBuf) -> Self {
        file_reader(&mut self.ram, file);
        self
    }

    fn draw(&mut self, frame: &mut [u8], dimensions: (u32, u32), tx: &SyncSender<f32>) -> Self {
        if self.sound_timer != 0 {
            self.sound_timer -= 1;
            tx.send(0.5).unwrap();
        } else {
            tx.send(0.0).unwrap();
        }
        if self.delay_timer != 0 {
            self.delay_timer -= 1;
        }
        for(i, pixel) in frame.chunks_exact_mut(4).enumerate() {
            let x = i % dimensions.0 as usize;
            let y = i / dimensions.0 as usize;

            let sx = (x * 64) / dimensions.0 as usize;
            let sy = (y * 32) / dimensions.1 as usize;

            let rgba = if get_bit(self.display[sy], sx as u8) {
                [0xff, 0xff, 0xff, 0xff]
            } else {
                [0x00, 0x00, 0x00, 0xff]
            };

            pixel.copy_from_slice(&rgba);
        }
        self.display_wait = false;
        //info!("done");
        *self
    }

    fn update(&mut self, instructions: u32) -> Self {
        for i in 0..instructions {
            let pointer = self.pc as usize;
            let opcodes: [u8; 4] = [
                ((self.ram[pointer] & 240) / 16),
                (self.ram[pointer] & 15),
                ((self.ram[pointer + 1] & 240) /16),
                (self.ram[pointer + 1] & 15)
            ];
            self.pc += 2;
            //info!("{:?}",opcodes);
            match opcodes[0] {
                0x00 => {
                    match opcodes[1] {
                        0x00 => {
                            match opcodes[2] {
                                0x0e => {
                                    match opcodes[3] {
                                        0x00 => {
                                            self.clear_screen();
                                        },
                                        0x0e => {
                                            if self.stack_pointer != 0 {
                                                self.pc = self.stack[self.stack_pointer as usize];
                                                self.stack_pointer -= 1;
                                            } else {
                                                error!("at root subroutine")
                                            }
                                        },
                                        _ => {
                                            error!("Does not support machine code");
                                        }
                                    }
                                },
                                _ => {
                                    error!("Does not support machine code");
                                }
                            }
                        },
                        _ => {
                            error!("Does not support machine code");
                        }
                    }
                },
                0x01 => {
                    let position = (opcodes[1] as u16) << 8 | (opcodes[2] as u16) << 4 | opcodes[3] as u16;
                    self.pc = position;
                },
                0x02 => {
                    if self.stack_pointer != 63 {
                        self.stack_pointer += 1;
                        self.stack[self.stack_pointer as usize] = self.pc;
                        self.pc = ((opcodes[1] as u16) << 8) | ((opcodes[2] as u16) << 4) | opcodes[3] as u16;
                    } else {
                        error!("reached the limit of the stack")
                    }
                },
                0x03 => {
                    //info!("{},{},{}",(self.registers[opcodes[1] as usize]),(((opcodes[2]) << 4) + opcodes[3]),(self.registers[opcodes[1] as usize]) == (((opcodes[2]) << 4) + opcodes[3]));
                    if (self.registers[opcodes[1] as usize]) == (((opcodes[2]) << 4) + opcodes[3]) {
                        self.pc += 2;
                    }
                },
                0x04 => {
                    if (self.registers[opcodes[1] as usize]) != (((opcodes[2]) << 4) + opcodes[3]) {
                        self.pc += 2;
                    }
                },
                0x05 => {
                    match opcodes[3] {
                        0x00 => {
                            if self.registers[opcodes[1] as usize] == self.registers[opcodes[2] as usize] {
                                self.pc += 2;
                            }
                        },
                        _ => {error!("unknown opcode")}
                    }
                },
                0x06 => {
                    self.registers[opcodes[1] as usize] = opcodes[2] << 4 | opcodes[3];
                },
                0x07 => {
                    let value = (opcodes[2] << 4 | opcodes[3]) as u16;
                    self.registers[opcodes[1] as usize] = (((self.registers[opcodes[1] as usize] as u16) + value) & 255) as u8;
                },
                0x08 => {
                    match opcodes[3] {
                        0x00 => {
                            self.registers[opcodes[1] as usize] = self.registers[opcodes[2] as usize];
                        },
                        0x01 => {
                            self.registers[opcodes[1] as usize] = self.registers[opcodes[1] as usize] | self.registers[opcodes[2] as usize];
                            self.registers[0x0f] = 0;
                        },
                        0x02 => {
                            self.registers[opcodes[1] as usize] = self.registers[opcodes[1] as usize] & self.registers[opcodes[2] as usize];
                            self.registers[0x0f] = 0;
                        },
                        0x03 => {
                            self.registers[opcodes[1] as usize] = self.registers[opcodes[1] as usize] ^ self.registers[opcodes[2] as usize];
                            self.registers[0x0f] = 0;
                        },
                        0x04 => {
                            let added_base = (self.registers[opcodes[1] as usize] as u16) + (self.registers[opcodes[2] as usize] as u16);
                            self.registers[opcodes[1] as usize] = (added_base & 255) as u8;
                            if added_base > 255 {
                                self.registers[0x0f] = 1;
                            } else {
                                self.registers[0x0f] = 0;
                            }
                        },
                        0x05 => {
                            let r1 = self.registers[opcodes[1] as usize];
                            let r2 = self.registers[opcodes[2] as usize];
                            //info!("{},{}", r1 , r2);
                            let no_underflow = (r1 >= r2);
                            if no_underflow {
                                //info!("{}",r1 - r2);
                                self.registers[opcodes[1] as usize] = r1 - r2;
                            } else {
                                self.registers[opcodes[1] as usize] = 255 - ((r2 - r1) - 1);
                            }
                            self.registers[0x0f] =  no_underflow as u8;
                        },
                        0x06 => {
                            if MODERNSHIFT {
                                let t = ((self.registers[opcodes[1] as usize] & 1) != 0) as u8;
                                self.registers[opcodes[1] as usize] = self.registers[opcodes[1] as usize] >> 1;
                                self.registers[0x0f] = t;
                            } else {
                                self.registers[opcodes[1] as usize] = self.registers[opcodes[2] as usize];
                                let t = ((self.registers[opcodes[1] as usize] & 1) != 0) as u8;
                                self.registers[opcodes[1] as usize] = self.registers[opcodes[1] as usize] >> 1;
                                self.registers[0x0f] = t;
                            }
                        },
                        0x07 => {
                            let r1 = self.registers[opcodes[1] as usize];
                            let r2 = self.registers[opcodes[2] as usize];
                            let no_underflow = (r2 >= r1);
                            //self.registers[0x0f] =  no_underflow as u8;
                            if no_underflow {
                                self.registers[opcodes[1] as usize] = r2 - r1;
                            } else {
                                self.registers[opcodes[1] as usize] = 255 - ((r1 - r2) - 1);
                            }
                            self.registers[0x0f] =  no_underflow as u8;
                        },
                        0x0e => {
                            if MODERNSHIFT {
                                let t = ((self.registers[opcodes[1] as usize] & 128) != 0) as u8;
                                self.registers[opcodes[1] as usize] = self.registers[opcodes[1] as usize] << 1;
                                self.registers[0x0f] = t;
                            } else {
                                self.registers[opcodes[1] as usize] = self.registers[opcodes[2] as usize];
                                let t = ((self.registers[opcodes[1] as usize] & 128) != 0) as u8;
                                self.registers[opcodes[1] as usize] = self.registers[opcodes[1] as usize] << 1;
                                self.registers[0x0f] = t;
                            }
                        },
                        _ => {error!("unknown opcode")}
                    }
                },
                0x09 => {
                    match opcodes[3] {
                        0x00 => {
                            if self.registers[opcodes[1] as usize] != self.registers[opcodes[2] as usize] {
                                self.pc += 2;
                            }
                        },
                        _ => {error!("unknown opcode")}
                    }
                },
                0x0a => {
                    self.index_register = (opcodes[1] as u16) << 8 | (opcodes[2] as u16) << 4 | (opcodes[3] as u16);
                },
                0x0b => {
                    if OLDJUMP {
                        self.pc = ((opcodes[1] as u16) << 8) + ((opcodes[2] as u16) << 4) + (opcodes[3] as u16) + (self.registers[0] as u16);
                    } else {
                        self.pc = ((opcodes[1] as u16) << 8) + ((opcodes[2] as u16) << 4) + (opcodes[3] as u16) + (self.registers[opcodes[1] as usize] as u16)
                    }
                },
                0x0c => {
                    let mut rng = rand::rng();
                    let random = rng.random::<u8>();
                    self.registers[opcodes[1] as usize] = random & (opcodes[2] << 4 | opcodes[3]);
                },
                0x0d => {
                    if !self.display_wait {
                        let x = self.registers[opcodes[1] as usize] % 64;
                        let y = self.registers[opcodes[2] as usize] % 32;
                        self.registers[0x0f] = 0;
                        for n in 0..opcodes[3] {
                            let spritedata = self.ram[self.index_register as usize + n as usize];
                            let mut checkdata: u8 = 0;
                            for i in 0..8 {
                                let bit = get_bit_u8(spritedata, i);
                                checkdata += (bit as u8) << (7 - i);
                                if (x+i <= 63) && (y+n <= 31)  {
                                    if (get_bit(self.display[(y+n) as usize], (x+i)) && bit) {
                                        self.registers[0x0f] = 1;
                                    }
                                    self.display[(y + n) as usize] = self.display[(y + n) as usize] ^ ((bit as u64) << (x + i));
                                }
                            }
                        //info!("{},{}", spritedata, checkdata);
                        }
                        self.display_wait = true;
                    } else {
                        self.pc -= 2;
                    }
                },
                0x0e => {
                    match opcodes[2] {
                        0x09 => {
                            match opcodes[3] {
                                0x0e => {
                                    //info!("23: {}", self.registers[opcodes[1] as usize]);
                                    if self.registers[opcodes[1] as usize] < 16{
                                        if self.keys[(self.registers[opcodes[1] as usize] & 15) as usize]  {
                                            self.pc += 2;
                                        }
                                    } else {
                                        error!("invalid keycode");
                                    }
                                }
                                _ => {}
                            }
                        },
                        0x0a => {
                            match opcodes[3] {
                                0x01 => {
                                    //info!("a1 {}", self.registers[opcodes[1] as usize]);
                                    if self.registers[opcodes[1] as usize] < 16{
                                        if !self.keys[(self.registers[opcodes[1] as usize] & 15) as usize]  {
                                            self.pc += 2;
                                        }
                                    } else {
                                        error!("invalid keycode");
                                    }
                                },
                                _ => {error!("invalid opcode");}
                            }
                        },
                        _ => {error!("unknown opcode")}
                    }
                },
                0x0f => {
                    match opcodes[2] {
                        0x00 => {
                            match opcodes[3] {
                                0x07 => {
                                    self.registers[opcodes[1] as usize] = self.delay_timer;
                                },
                                0x0a => {
                                    let mut keypress = false;
                                    if !self.get_key_pause {
                                        self.get_key_pause = true;
                                        for i in 0..16 {
                                            self.keys_released[i] = !self.keys[i];
                                            self.keys_pressed[i] = false;
                                        }
                                    } else {
                                        for i in 0..16 {
                                            if !keypress {
                                                if self.keys_pressed[i] && !self.keys[i] {
                                                    self.registers[opcodes[1] as usize] = i as u8;
                                                    self.get_key_pause = false;
                                                    keypress = true;
                                                } else if self.keys_released[i] && self.keys[i] {
                                                    self.keys_pressed[i] = true;
                                                } else if !self.keys_released[i] && !self.keys[i] {
                                                    self.keys_released[i] = true;
                                                }
                                            }
                                        }
                                    }
                                    if !keypress {
                                        self.pc -= 2;
                                    }
                                },
                                _ => {
                                    error!("unknown opcode");
                                }
                            }
                        }
                        0x01 => {
                            match opcodes[3] {
                                0x05 => {
                                    self.delay_timer = self.registers[opcodes[1] as usize];
                                },
                                0x08 => {
                                    self.sound_timer = self.registers[opcodes[1] as usize];
                                }
                                0x0e => {
                                    if self.index_register + self.registers[opcodes[1] as usize] as u16 >= 0x1000 {
                                        self.registers[0x0f] = 1;
                                    }
                                    self.index_register = (self.index_register + self.registers[opcodes[1] as usize] as u16) & 0x0FFF;
                                },
                                _ => {error!("unknown opcode");}
                            }
                        },
                        0x02 => {
                            match opcodes[3] {
                                0x09 => {
                                    self.index_register = 0x050 + (((self.registers[opcodes[1] as usize] & 0x0F) as u16) * 5);
                                },
                                _ => {error!("unknown opcode");}
                            }
                        },
                        0x03 => {
                            match opcodes[3] {
                                0x03 => {
                                    self.ram[self.index_register as usize] = self.registers[opcodes[1] as usize] / 100;
                                    self.ram[self.index_register as usize + 1] = (self.registers[opcodes[1] as usize] / 10) % 10;
                                    self.ram[self.index_register as usize + 2] = self.registers[opcodes[1] as usize] % 10;
                                },
                                _ => {error!("unknown opcode");}
                            }
                        }
                        0x05 => {
                            match opcodes[3] {
                                0x05 => {
                                    if MODERNLOADSAVE {
                                        for i in 0..=opcodes[1] {
                                            self.ram[(self.index_register + i as u16) as usize] = self.registers[i as usize];
                                        }
                                    } else {
                                        for i in 0..=opcodes[1] {
                                            //self.index_register += i as u16;
                                            self.ram[self.index_register as usize] = self.registers[i as usize];
                                            self.index_register += 1;
                                        }
                                    }
                                },
                                _ => {error!("unknown opcode");}
                            }
                        },
                        0x06 => {
                            match opcodes[3] {
                                0x05 => {
                                    if MODERNLOADSAVE {
                                        for i in 0..=opcodes[1] {
                                            self.registers[i as usize] = self.ram[(self.index_register + i as u16) as usize];
                                        }
                                    } else {
                                        for i in 0..=opcodes[1] {
                                            //self.index_register += 1;
                                            self.registers[i as usize] = self.ram[self.index_register as usize];
                                            self.index_register += 1;
                                        }
                                    }
                                },
                                _ => {error!("unknown opcode");}
                            }
                        },
                        _ => {error!("unknown opcode")}
                    }

                },
                _ => {
                    error!("out of scope opcode");
                }
            }
        }
        //info!("{:?}",self.display);
        *self
    }

    fn clear_screen(&mut self) -> Self {
        self.display = [0;32];
        *self
    }
}

fn main() {
    tracing_subscriber::registry()
        .with(
            tracing_subscriber::EnvFilter::try_from_default_env()
                .unwrap_or_else(|_| "INFO=debug".into()),
        )
        .with(tracing_subscriber::fmt::layer())
        .init();

    let args = Args::parse();
    let mut chip8 = Chip8::new().insert_font().load_file(args.file);
    let (tx, rx) = sync_channel(1);

    println!("ram: {:?}", chip8.ram);

    let event_loop = EventLoop::new();

    let window = { let window = WindowBuilder::new()
        .with_title("Chip8 Emulator")
        .with_inner_size(LogicalSize::new(320, 240))
        .with_min_inner_size(LogicalSize::new(600, 300))
        .build(&event_loop)
        .unwrap();
        Arc::new(window)
    };

    let mut pixels = {
        let window_size = window.inner_size();
        let surface_texture = SurfaceTexture::new(window_size.width, window_size.height, Arc::clone(&window));
        Pixels::new(window_size.width, window_size.height, surface_texture).unwrap()
    };

    let mut audio = AudioPlayer::make(rx);

    audio.run();

    let test: u64 = 1;
    //info!("{}", get_bit(test, 0));
    let mut start = Instant::now();
    event_loop.run(move |event, _, control_flow| {
        match event {
            Event::WindowEvent {event, ..} => match event {
                WindowEvent::CloseRequested => *control_flow = ControlFlow::Exit,
                WindowEvent::Resized(size) => {
                    if let Err(e) = pixels.resize_surface(size.width, size.height) {
                        error!("pixels.resize_surface");
                        *control_flow = ControlFlow::Exit;
                    }
                },
                WindowEvent::KeyboardInput {
                    event:
                        KeyEvent {
                            logical_key,
                            state,
                            ..
                        },
                    ..
                } => {
                    match logical_key {
                        Key::Character("1") => {
                            set_keystate(&mut chip8.keys, 1, state);
                        },
                        Key::Character("2") => {
                            set_keystate(&mut chip8.keys, 2, state);
                        },
                        Key::Character("3") => {
                            set_keystate(&mut chip8.keys, 3, state);
                        },
                        Key::Character("4") => {
                            set_keystate(&mut chip8.keys, 0x0c, state);
                        },
                        Key::Character("q") => {
                            set_keystate(&mut chip8.keys, 4, state);
                        },
                        Key::Character("w") => {
                            set_keystate(&mut chip8.keys, 5, state);
                        },
                        Key::Character("e") => {
                            set_keystate(&mut chip8.keys, 6, state);
                        },
                        Key::Character("r") => {
                            set_keystate(&mut chip8.keys, 0x0d, state);
                        },
                        Key::Character("a") => {
                            set_keystate(&mut chip8.keys, 7, state);
                        },
                        Key::Character("s") => {
                            set_keystate(&mut chip8.keys, 8, state);
                        },
                        Key::Character("d") => {
                            set_keystate(&mut chip8.keys, 9, state);
                        },
                        Key::Character("f") => {
                            set_keystate(&mut chip8.keys, 0x0e, state);
                        },
                        Key::Character("z") => {
                            set_keystate(&mut chip8.keys, 0x0a, state);
                        },
                        Key::Character("x") => {
                            set_keystate(&mut chip8.keys, 0, state);
                        },
                        Key::Character("c") => {
                            set_keystate(&mut chip8.keys, 0x0b, state);
                        },
                        Key::Character("v") => {
                            set_keystate(&mut chip8.keys, 0x0f, state);
                        },
                        _ => {}
                    }
                }
                _ => {}
            },
            Event::MainEventsCleared => {
                //info!(chip8.pc);
                chip8.update(40);
                //info!(chip8.pc);
                //for i in chip8.display {
                //    println!("{:#64b}", i);
                //}
                window.request_redraw();
            }
            Event::RedrawRequested(_) => {
                chip8.draw(pixels.frame_mut(), (window.inner_size().width, window.inner_size().height), &tx);
                if let Err(err) = pixels.render() {
                    error!("pixels.render");
                    *control_flow = ControlFlow::Exit;
                }
                //let duration = start.elapsed();
                //info!("{:?}", duration);
                //start = Instant::now();
            }
            _ => {}
        }
    });


}

fn file_reader(ram: &mut [u8; 4096], file: PathBuf) {
    let mut file = File::open(file).unwrap();
    let mut file_contents:Vec<u8> = Vec::new();
    file.read_to_end(&mut file_contents).unwrap();
    for i in 0..file_contents.len() {
        ram[i + 0x200] = file_contents[i];
    }
}

fn font_initialiser(ram: &mut [u8; 4096]) {
    let font: [u8; 80] = [
        0xF0, 0x90, 0x90, 0x90, 0xF0, // 0
        0x20, 0x60, 0x20, 0x20, 0x70, // 1
        0xF0, 0x10, 0xF0, 0x80, 0xF0, // 2
        0xF0, 0x10, 0xF0, 0x10, 0xF0, // 3
        0x90, 0x90, 0xF0, 0x10, 0x10, // 4
        0xF0, 0x80, 0xF0, 0x10, 0xF0, // 5
        0xF0, 0x80, 0xF0, 0x90, 0xF0, // 6
        0xF0, 0x10, 0x20, 0x40, 0x40, // 7
        0xF0, 0x90, 0xF0, 0x90, 0xF0, // 8
        0xF0, 0x90, 0xF0, 0x10, 0xF0, // 9
        0xF0, 0x90, 0xF0, 0x90, 0x90, // A
        0xE0, 0x90, 0xE0, 0x90, 0xE0, // B
        0xF0, 0x80, 0x80, 0x80, 0xF0, // C
        0xE0, 0x90, 0x90, 0x90, 0xE0, // D
        0xF0, 0x80, 0xF0, 0x80, 0xF0, // E
        0xF0, 0x80, 0xF0, 0x80, 0x80  // F
    ];

    for i in 0..80 {
        ram[i + 0x050] = font[i];
    }
}



fn get_bit(num: u64, bit: u8) -> bool {
    //info!(bit);
    if bit <= 63 {
        let mask: u64 = 1 << bit;
        let bitvalue = ((num & mask)) != 0;
        bitvalue
    } else {
        false
    }
}

fn get_bit_u8(num: u8, bit: u8) -> bool {
    if bit <= 8 {
        let mask: u8 = 1 << (7 - bit);
        let bitvalue = (num & mask) != 0;
        //info!("{},{},{},{}", num, bit, mask, bitvalue);
        bitvalue
    } else {
        false
    }
}

fn set_keystate(keys: &mut [bool; 16], key: u8, state: ElementState) {
    if state == ElementState::Pressed {
        keys[key as usize] = true;
    } else if state == ElementState::Released {
        keys[key as usize] = false;
    }
}